@@ -310,6 +310,7 @@ public:
 
   void ReplaceOneCharAtPos(unsigned pos, char c) { _chars[pos] = c; }
 
+  char *GetBuf() { return _chars; }
   /* GetBuf(minLen): provides the buffer that can store
      at least (minLen) characters and additional null terminator.
      9.35: GetBuf doesn't preserve old characters and terminator */
@@ -548,6 +549,8 @@ public:
 
   void ReplaceOneCharAtPos(unsigned pos, wchar_t c) { _chars[pos] = c; }
 
+  wchar_t *GetBuf() { return _chars; }
+
   wchar_t *GetBuf(unsigned minLen)
   {
     if (minLen > _limit)
@@ -574,7 +577,7 @@ public:
   }
 
   UString &operator=(wchar_t c);
-  UString &operator=(char c) { return (*this)=((wchar_t)c); }
+  UString &operator=(char c) { return (*this)=((wchar_t)(unsigned char)c); }
   UString &operator=(const wchar_t *s);
   UString &operator=(const UString &s);
   void SetFrom(const wchar_t *s, unsigned len); // no check
